use std::marker::PhantomData;

use super::*;

pub trait ParseMode {
    type Mode: ?Sized;
}

pub trait WithParseMode: ParseMode {
    type WithMode: ?Sized;
}

impl<T: ?Sized + ParseMode> WithParseMode for T {
    type WithMode = WithMode<Self, <Self as ParseMode>::Mode>;
}

pub struct WithMode<T: ?Sized, M: ?Sized>(PhantomData<M>, T);

pub trait FactoryProxy<'a, Ctx: Context<'a>> {
    type F: FactoryBase<'a, Ctx> + ParseMode;

    fn pdeserialize(f: &Self::F, inctx: impl InCtx<'a, Ctx>) -> ParseResult<'a, Ctx, Self::F>;
    fn pextend(
        f: &Self::F,
        mentionable: Mtbl<'a, Ctx, Self::F>,
        tail: &[u8],
    ) -> ParseResult<'a, Ctx, Self::F>;
}

impl<'a, Ctx: Context<'a>, F: FactoryBase<'a, Ctx> + WithParseMode> ParseFactory<'a, Ctx> for F
where
    F::WithMode: FactoryProxy<'a, Ctx, F = Self>,
{
    fn deserialize(&self, inctx: impl InCtx<'a, Ctx>) -> ParseResult<'a, Ctx, Self> {
        <F::WithMode as FactoryProxy<'a, Ctx>>::pdeserialize(self, inctx)
    }

    fn extend(&self, mentionable: Self::Mtbl, tail: &[u8]) -> ParseResult<'a, Ctx, Self> {
        <F::WithMode as FactoryProxy<'a, Ctx>>::pextend(self, mentionable, tail)
    }
}

impl<'a, Ctx: Context<'a>, F: ParseFactory<'a, Ctx> + WithParseMode> Factory<'a, Ctx> for F
where
    F::Mtbl: MentionableTop<'a, Ctx>,
{
    type _Mtbl = Self::Mtbl;
}

pub struct RegularMode;

pub trait RegularFactory<'a, Ctx: Context<'a>>:
    FactoryBase<'a, Ctx> + ParseMode<Mode = RegularMode>
{
    fn rdeserialize(&self, inctx: impl InCtx<'a, Ctx>) -> ParseResult<'a, Ctx, Self>;
    fn rextend(&self, mentionable: Self::Mtbl, tail: &[u8]) -> ParseResult<'a, Ctx, Self>;
}

impl<'a, Ctx: Context<'a>, F: RegularFactory<'a, Ctx>> FactoryProxy<'a, Ctx>
    for WithMode<F, RegularMode>
{
    type F = F;

    fn pdeserialize(f: &Self::F, inctx: impl InCtx<'a, Ctx>) -> ParseResult<'a, Ctx, Self::F> {
        f.rdeserialize(inctx)
    }

    fn pextend(
        f: &Self::F,
        mentionable: Mtbl<'a, Ctx, Self::F>,
        tail: &[u8],
    ) -> ParseResult<'a, Ctx, Self::F> {
        f.rextend(mentionable, tail)
    }
}
